// SPDX-License-Identifier: GPL-3.0-or-later

#pragma once


#include "platform/platformnativeinterface.h"

#include <QKeyEvent>
#include <QString>

class DummyPlatform : public PlatformNativeInterface
{
public:
    PlatformWindowPtr getWindow(WId) override { return PlatformWindowPtr(); }

    PlatformWindowPtr getCurrentWindow() override { return PlatformWindowPtr(); }

    bool canGetWindowTitle() override { return false; }

    bool canAutostart() override { return false; }

    bool isAutostartEnabled() override { return false; }

    void setAutostartEnabled(bool) override {}

    bool setPreventScreenCapture(WId, bool) override { return false; }
    bool canPreventScreenCapture() override { return false; }

    QCoreApplication *createConsoleApplication(int &argc, char **argv) override;

    QApplication *createServerApplication(int &argc, char **argv) override;

    QGuiApplication *createClipboardProviderApplication(int &argc, char **argv) override;

    QCoreApplication *createClientApplication(int &argc, char **argv) override;

    QGuiApplication *createTestApplication(int &argc, char **argv) override;

    PlatformClipboardPtr clipboard() override;

    QStringList getCommandLineArguments(int argc, char **argv) override;

    bool findPluginDir(QDir *pluginsDir) override;

    QString defaultEditorCommand() override;

    QString translationPrefix() override;

    QString themePrefix() override { return QString(); }

    qint64 processResidentMemoryBytes(qint64) override { return -1; }
};
