// SPDX-License-Identifier: GPL-3.0-or-later

#include "mimetypes.h"

#include <QLatin1String>

const QLatin1String mimePrivatePrefix(COPYQ_MIME_PRIVATE_PREFIX);
const QLatin1String mimeText("text/plain");
const QLatin1String mimeTextUtf8("text/plain;charset=utf-8");
const QLatin1String mimeHtml("text/html");
const QLatin1String mimeUriList("text/uri-list");
const QLatin1String mimeWindowTitle(COPYQ_MIME_PREFIX "owner-window-title");
const QLatin1String mimeItems(COPYQ_MIME_PREFIX "item");
const QLatin1String mimeItemNotes(COPYQ_MIME_PREFIX "item-notes");
const QLatin1String mimeIcon(COPYQ_MIME_PREFIX "item-icon");
const QLatin1String mimeOwner(COPYQ_MIME_PREFIX "owner");
const QLatin1String mimeClipboardMode(COPYQ_MIME_PREFIX "clipboard-mode");
const QLatin1String mimeCurrentTab(COPYQ_MIME_PREFIX "current-tab");
const QLatin1String mimeSelectedItems(COPYQ_MIME_PREFIX "selected-items");
const QLatin1String mimeCurrentItem(COPYQ_MIME_PREFIX "current-item");
const QLatin1String mimeHidden(COPYQ_MIME_PREFIX "hidden");
const QLatin1String mimeShortcut(COPYQ_MIME_PREFIX "shortcut");
const QLatin1String mimeColor(COPYQ_MIME_PREFIX "color");
const QLatin1String mimeOutputTab(COPYQ_MIME_PREFIX "output-tab");
const QLatin1String mimeDisplayItemInMenu(COPYQ_MIME_PREFIX "display-item-in-menu");
const QLatin1String mimeSecret(COPYQ_MIME_PREFIX "secret");
