// SPDX-License-Identifier: GPL-3.0-or-later

#pragma once

class QByteArray;
class QPoint;
class QString;
class QVariant;
class QWidget;

QVariant geometryOptionValue(const QString &optionName);
void setGeometryOptionValue(const QString &optionName, const QVariant &value);

void restoreWindowGeometry(QWidget *w, bool openOnCurrentScreen);

void saveWindowGeometry(QWidget *w, bool openOnCurrentScreen);

QByteArray mainWindowState(const QString &mainWindowObjectName);

void saveMainWindowState(const QString &mainWindowObjectName, const QByteArray &state);

void moveToCurrentWorkspace(QWidget *w);

void moveWindowOnScreen(QWidget *w, QPoint pos);

void setGeometryGuardBlockedUntilHidden(QWidget *w, bool blocked);

void ensureWindowOnScreen(QWidget *w);
bool isGeometryGuardBlockedUntilHidden(const QWidget *w);
