// SPDX-License-Identifier: GPL-3.0-or-later
#pragma once

namespace ScriptOverrides {
enum ScriptOverrides {
    Paste = 0,
    OnItemsAdded = 1,
    OnItemsRemoved = 2,
    OnItemsChanged = 3,
    OnTabSelected = 4,
    OnItemsLoaded = 5,
    OnExit = 6,
};
}
