// SPDX-License-Identifier: GPL-3.0-or-later
// Compile miniaudio implementation as part of CopyQ.
// The miniaudio.h header is provided by the system (e.g. miniaudio-devel).
#define MINIAUDIO_IMPLEMENTATION
#include <miniaudio.h>
